; *******************************************************
; *                                                     *
; *     Turbo Pascal Run-time Library                   *
; *     Main module                                     *
; *                                                     *
; *******************************************************

        TITLE   MAIN

        INCLUDE SE.ASM

DATA    SEGMENT WORD PUBLIC
; Externals
        EXTRN   OvrHeapSize:WORD,OvrHeapOrg:WORD,OvrHeapPtr:WORD
        EXTRN   OvrHeapEnd:WORD,OvrLoadList:WORD,HeapOrg:DWORD
        EXTRN   HeapPtr:DWORD,HeapEnd:DWORD,FreeList:DWORD
        EXTRN   HeapError:DWORD,ExitProc:DWORD,ExitCode:WORD
        EXTRN   ErrorAddr:DWORD,PrefixSeg:WORD,InOutRes:WORD
        EXTRN   SaveInt00:DWORD
        EXTRN   Test8086:BYTE
        extrn   Modex:byte,ErrAddr:dword
        extrn Terminator:dword,SaveIntTab:byte
DATA    ENDS

CODE    SEGMENT BYTE PUBLIC

        ASSUME  CS:CODE,DS:DATA

; Externals
        EXTRN   ClearDSeg:NEAR

; Publics
        PUBLIC  InitTurbo,HaltTurbo,HaltError;,PrintString
        PUBLIC  Terminate,LibEntry,LibExit,LibExitProc,InitTask

        public ExecSwap

ExecSwap:
        arg FileName,dword,1
        arg Params,dword,1
        entry far

        exit

; Initialize runtime library. First instruction in any program
; is a call to this routine.
InitTurbo:
        MOV     DX,DATA                 ;Initialize DS
        MOV     DS,DX
        MOV     PrefixSeg,ES            ;Save PSP segment
        XOR     BP,BP                   ;End of stack frame chain             ; .ASM <- that is for BP
        CALL    ClearDSeg               ;Zero fill data segment
        CALL    Check8086               ;Determine CPU type
        MOV     AX,SP                   ;Compute first free segment
        ADD     AX,4+15                 ;address in AX
        MOV     CL,4
        SHR     AX,CL
        MOV     DX,SS
        ADD     AX,DX
        MOV     OvrHeapOrg,AX           ;Initialize overlay manager
        MOV     OvrHeapPtr,AX           ;variables
        ADD     AX,OvrHeapSize
        MOV     OvrHeapEnd,AX
        MOV     HeapOrg.seg,AX          ;Initialize heap manager
        MOV     HeapPtr.seg,AX          ;variables
        MOV     FreeList.seg,AX
        MOV     ES,PrefixSeg
        MOV     AX,ES:pspMemTop
        MOV     HeapEnd.seg,AX
        MOV     HeapError.ofs,OFFSET CS:HeapFailure
        MOV     HeapError.seg,CS

        MOV     DI,OFFSET SaveInt00     ;Capture interrupt vectors
        MOV     SI,OFFSET SaveIntTab
        cld
        lodsb
        xor ch, ch
        mov cl, al
@@1:    lodsb
        mov ah, dosGetInt
        int DOS
        mov [di].ofs,BX
        mov [di].seg,ES
        add di, 4
        loop @@1
        PUSH    DS                      ;Install interrupt handlers
         PUSH    CS
         POP     DS
         MOV     DX,OFFSET Int00Handler
         MOV     AX,dosSetInt*256+00H
         INT     DOS
         MOV     DX,OFFSET Int06Handler
         MOV     AX,dosSetInt*256+06H
         INT     DOS
         MOV     DX,OFFSET Int23Handler
         MOV     AX,dosSetInt*256+23H
         INT     DOS
         MOV     DX,OFFSET Int24Handler
         MOV     AX,dosSetInt*256+24H
         INT     DOS
         MOV     DX,OFFSET Int3FHandler
         MOV     AX,dosSetInt*256+3FH
         INT     DOS
        POP     DS
        RETF                            ;Back to main program

; CPU detection routine. Sets the Test8086 variable according to the
; type of CPU detected.
Check8086:
        XOR     AX,AX                   ;0 means 8088/8086
        PUSHF                           ;BX = Flags
        POP     BX
        AND     BH,0FH                  ;Clear bits 12-15
        PUSH    BX                      ;Flags = BX
        POPF
        PUSHF                           ;CX = Flags
        POP     CX
        AND     CH,0F0H                 ;Bits 12-15 set?
        CMP     CH,0F0H
        JE      @@1                     ;Yes, 8086
        INC     AX                      ;1 means 80286
        OR      BH,0F0H                 ;Set bits 12-15
        PUSH    BX                      ;Flags = BX
        POPF
        PUSHF                           ;CX = Flags
        POP     CX
        AND     CH,0F0H                 ;Bits 12-15 cleared?
        JE      @@1                     ;Yes, 80286
        INC     AX                      ;2 means 80386
@@1:    MOV     Test8086,AL             ;Save CPU test result
        RET

; Default heap error handler. Return 0 to indicate run-time error.
HeapFailure:
        XOR     AX,AX
        RETF    2

; Critical error interrupt handler. Control arrives here when
; DOS encounters a "critical error". Convert critical error code
; to Turbo Pascal I/O error code, and return.
Int24Handler:
        STI                             ;Enable interrupts
        ADD     SP,6                    ;Remove INT 24H return info
        POP     AX                      ;Get INT 21H AX register
        AND     DI,01FH                 ;Return AX=150..181
        ADD     DI,150
        CMP     AH,39H                  ;DOS 2.0 style function?
        JAE     @@1                     ;Yes, @@1
        MOV     DI,0FFFFH               ;Return AX=0FFFFH
@@1:    PUSH    DI                      ;Save error code
        MOV     AH,54H                  ;Dummy function call to get
        INT     DOS                     ;DOS into a stable state
        MOV     BP,SP                   ;Set CF in return flags
        OR      BYTE PTR [BP+22],1
        POP     AX                      ;Restore registers
        POP     BX
        POP     CX
        POP     DX
        POP     SI
        POP     DI
        POP     BP
        POP     DS
        POP     ES
        IRET                            ;Return to caller

; Overlay interrupt handler. Control arrives here if the program
; calls an overlay procedure before initializing the overlay manager.
Int3FHandler:
        MOV     AX,208
        ADD     SP,6
        JMP     SHORT HaltError

; Divide by zero interrupt handler. Control arrives here upon
; executing a DIV or IDIV instruction with a zero divisor.
Int00Handler:
        MOV     AX,200
        jmp short HaltError

; Exceptions' 6 and 13 handler. Control arrives here upon
; executing a invalid instruction.
Int06Handler:
@@1:    mov ax, 216

; RunError standard procedure
HaltError:
        POP     CX
        POP     BX
        jmp short Terminate

; Dummy entry points
LibEntry:
LibExit:
LibExitProc:
InitTask:

; Control-C interrupt handler. Control arrives here when DOS
; detects a Ctrl-C or Ctrl-Break.
Int23Handler:
        MOV     AX,255

; Halt standard procedure
HaltTurbo:
        XOR     CX,CX
        XOR     BX,BX

; Terminate program and return to DOS
; In    AX    = Exit code
;       BX:CX = Error address (or NIL)
Terminate:
        MOV DX,DATA                 ;Reset DS
        MOV DS,DX
        STI                         ;Enable interrupts
        MOV ExitCode,AX             ;Save exit code
        mov ErrAddr.ofs, cx
        mov ErrAddr.seg, bx
        jmp dword ptr Terminator

CODE    ENDS

        END
